# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2019/05/13

"""State definition bases and mix-ins."""

import typing as T
import logging as lg

from .. import _util

_logger = lg.getLogger(__name__)
_default = _util.DefaultParameter()
STATES_ERRORS = (
    "ALL",
    "Timeout",
    "TaskFailed",
    "Permissions",
    "ResultPathMatchFailure",
    "ParameterPathFailure",
    "BranchFailed",
    "NoChoiceMatched")


class State:
    """Abstract state.

    Args:
        name: name of state
        comment: state description
        input_path: state input filter JSONPath, ``None`` for empty input
        output_path: state output filter JSONPath, ``None`` for discarded
            output
    """

    def __init__(
            self,
            name: str,
            comment: str = _default,
            input_path: T.Union[str, None] = _default,
            output_path: T.Union[str, None] = _default):
        self.name = name
        self.comment = comment
        self.input_path = input_path
        self.output_path = output_path

    def __str__(self):
        name = type(self).__name__
        return "%s [%s]" % (self.name, name)

    __repr__ = _util.easy_repr

    def add_to(self, states):
        """Add this state to a state-machine definition.

        Any child states will also be added to the definition.

        Arguments:
            states (dict[str, State]): state-machine states
        """

        if states.get(self.name, self) != self:
            raise ValueError("State name '%s' already registered" % self.name)
        states[self.name] = self

    def to_dict(self) -> T.Dict[str, _util.JSONable]:
        """Convert this state to a definition dictionary.

        Returns:
            definition
        """

        defn = {"Type": type(self).__name__}
        if self.comment != _default:
            defn["Comment"] = self.comment
        if self.input_path != _default:
            defn["InputPath"] = self.input_path
        if self.output_path != _default:
            defn["OutputPath"] = self.output_path
        return defn


class HasNext(State):
    """State able to advance mix-in.

    Args:
        name: name of state
        comment: state description
        input_path: state input filter JSONPath, ``None`` for empty input
        output_path: state output filter JSONPath, ``None`` for discarded
            output

    Attributes:
        next: next state to execute, or ``None`` if state is terminal
    """

    def __init__(
            self,
            name,
            comment=_default,
            input_path=_default,
            output_path=_default):
        super().__init__(
            name,
            comment=comment,
            input_path=input_path,
            output_path=output_path)
        self.next: T.Union[State, None] = None

    def add_to(self, states):
        super().add_to(states)
        if self.next is not None:
            self.next.add_to(states)

    def goes_to(self, state: State):
        """Set next state after this state finishes.

        Args:
            state: state to execute next
        """

        if self.next is not None:
            _logger.warning("Overriding current next state: %s" % self.next)
        self.next = state

    def remove_next(self):
        """Remove next state, making this state terminal."""
        self.next = None

    def to_dict(self):
        defn = super().to_dict()
        if self.next is None:
            defn["End"] = True
        else:
            defn["Next"] = self.next.name
        return defn


class HasResultPath(State):
    """State with result mix-in.

    Args:
        name: name of state
        comment: state description
        input_path: state input filter JSONPath, ``None`` for empty input
        output_path: state output filter JSONPath, ``None`` for discarded
            output
        result_path: task output location JSONPath, ``None`` for discarded
            output
    """

    def __init__(
            self,
            name,
            comment=_default,
            input_path=_default,
            output_path=_default,
            result_path: T.Union[str, None] = _default):
        super().__init__(
            name,
            comment=comment,
            input_path=input_path,
            output_path=output_path)
        self.result_path = result_path

    def to_dict(self):
        defn = super().to_dict()
        if self.result_path != _default:
            defn["ResultPath"] = self.result_path
        return defn


class CanRetry(State):
    """Retryable state mix-in.

    Args:
        name: name of state
        comment: state description
        input_path: state input filter JSONPath, ``None`` for empty input
        output_path: state output filter JSONPath, ``None`` for discarded
            output

    Attributes:
        retriers: error handler policies
    """

    def __init__(
            self,
            name,
            comment=_default,
            input_path=_default,
            output_path=_default):
        super().__init__(
            name,
            comment=comment,
            input_path=input_path,
            output_path=output_path)
        self.retriers: T.List[T.Tuple[T.Sequence[str], T.Dict[str, ...]]] = []

    def retry_for(
            self,
            errors: T.Sequence[str],
            interval: int = _default,
            max_attempts: int = _default,
            backoff_rate: float = _default):
        """Add a retry handler.

        Args:
            errors: codes of errors for retry to be executed. See AWS Step
                Functions documentation
            interval: (initial) retry interval (seconds)
            max_attempts: maximum number of attempts before re-raising error
            backoff_rate: retry interval increase factor between attempts
        """

        policy = {
            "interval": interval,
            "max_attempts": max_attempts,
            "backoff_rate": backoff_rate}
        self.retriers.append((errors, policy))

    @staticmethod
    def _retrier_defn(
            errors: T.Sequence[str],
            policy: T.Dict[str, T.Any]
    ) -> T.Dict[str, _util.JSONable]:
        """Build retry handler definition.

        Args:
            errors: codes of errors for retry handler to be invoked
            policy: retry handler policy

        Returns:
            definitions
        """

        _validate_errors(errors)
        defn = {"ErrorEquals": errors}
        if policy["interval"] != _default:
            defn["IntervalSeconds"] = policy["interval"]
        if policy["max_attempts"] != _default:
            defn["MaxAttempts"] = policy["max_attempts"]
        if policy["backoff_rate"] != _default:
            defn["BackoffRate"] = policy["backoff_rate"]
        return defn

    def _get_retrier_defns(self) -> T.List[T.Dict[str, _util.JSONable]]:
        """Build retry handler definitions.

        Returns:
            definitions
        """

        return [self._retrier_defn(e, p) for e, p in self.retriers]

    def to_dict(self):
        defn = super().to_dict()
        retry = self._get_retrier_defns()
        if retry:
            defn["Retry"] = retry
        return defn


class CanCatch(State):
    """Exception catching state mix-in.

    Args:
        name: name of state
        comment: state description
        input_path: state input filter JSONPath, ``None`` for empty input
        output_path: state output filter JSONPath, ``None`` for discarded
            output

    Attributes:
        catchers: error handler policies
    """

    def __init__(
            self,
            name,
            comment=_default,
            input_path=_default,
            output_path=_default):
        super().__init__(
            name,
            comment=comment,
            input_path=input_path,
            output_path=output_path)
        self.catchers: T.List[T.Tuple[T.Sequence[str], T.Dict[str, ...]]] = []

    def add_to(self, states):
        super().add_to(states)
        for _, policy in self.catchers:
            policy["next_state"].add_to(states)

    def catch(
            self,
            excs: T.Sequence[str],
            next_state: State,
            result_path: T.Union[str, None] = _default):
        """Add an error handler.

        Args:
            excs: errors for catch clause to be executed. If a string, must be
                one of the pre-defined errors (see AWS Step Functions
                documentation)
            next_state: state to execute for catch clause
            result_path: error details location JSONPath
        """

        if any(any(e in excs_ for e in excs) for excs_, _ in self.catchers):
            fmt = "Handler has already accounted-for exceptions: %s"
            _logger.warning(fmt % excs)
        policy = {"next_state": next_state, "result_path": result_path}
        self.catchers.append((excs, policy))

    @staticmethod
    def _catcher_defn(
            errors: T.Sequence[str],
            policy: T.Dict[str, T.Any]
    ) -> T.Dict[str, _util.JSONable]:
        """Build error handler definition.

        Args:
            errors: codes of errors for retry handler to be invoked
            policy: retry handler policy

        Returns:
            definitions
        """

        _validate_errors(errors)
        defn = {"ErrorEquals": errors, "Next": policy["next_state"].name}
        if policy["result_path"] != _default:
            defn["ResultPath"] = policy["result_path"]
        return defn

    def _get_catcher_defns(self) -> T.List[T.Dict[str, _util.JSONable]]:
        """Build error handler definitions.

        Returns:
            definitions
        """

        return [self._catcher_defn(e, p) for e, p in self.catchers]

    def to_dict(self):
        defn = super().to_dict()
        catch = self._get_catcher_defns()
        if catch:
            defn["Catch"] = catch
        return defn


def _validate_errors(errors: T.Sequence[str]):
    """Validate error conditions.

    Args:
        errors: condition error codes

    Raises:
        ValueError: invalid condition
    """

    if not errors:
        raise ValueError("Cannot have no-error condition")
    if "States.ALL" in errors and len(errors) > 1:
        msg = "Cannot combine 'States.ALL' condition with other errors"
        raise ValueError(msg)

    for err in errors:
        if err.startswith("States."):
            if err[7:] not in STATES_ERRORS:
                fmt = "States error name was '%s', must be one of: %s"
                raise ValueError(fmt % (err[7:], STATES_ERRORS))



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/07/11

"""Common utilities for ``sfini``."""

import inspect
import sys
import typing as T
import logging as lg
import functools as ft
from collections import abc

import boto3
from botocore import credentials
from botocore import client as botocore_client

_logger = lg.getLogger(__name__)
lg.getLogger("botocore").setLevel(lg.WARNING)
MAX_NAME_LENGTH = 79
INVALID_NAME_CHARACTERS = " \n\t<>{}[]?*\"#%\\^|~`$&,;:/"
DEBUG = "pytest" in sys.modules
JSONable = T.Union[None, bool, str, int, float, list, T.Dict[str, T.Any]]


class DefaultParameter:
    """Default parameter for step-functions definition."""
    def __bool__(self):
        return False

    def __eq__(self, other):
        return isinstance(self, type(other))

    def __str__(self):
        return "<unspecified>"

    def __repr__(self):
        return type(self).__name__ + "()"


def setup_logging(level: int = None):
    """Setup logging for ``sfini``, if logs would otherwise be ignored.

    Args:
        level: logging level (see ``logging``), default: leave unchanged
    """

    lg.basicConfig(
        format="%(asctime)s [%(levelname)8s] %(name)s: %(message)s",
        level=level)
    if level is not None:
        lg.getLogger().setLevel(level)
        [h.setLevel(level) for h in lg.getLogger().handlers]


def cached_property(fn: T.Callable) -> property:
    """Decorate a method as a cached property.

    The wrapped method's result is stored in the instance's ``__cache__``
    dictionary, with the method's name as key.

    Args:
        fn: method to decorate

    Returns:
        cached property
    """

    name = fn.__name__

    def _ensure_cache(self):
        if not hasattr(self, "__cache__"):
            self.__cache__ = {}

    @ft.wraps(fn)
    def wrapped(self):
        _ensure_cache(self)
        if name not in self.__cache__:
            self.__cache__[name] = fn(self)
        return self.__cache__[name]

    if DEBUG:  # for testing
        def fset(self, value):
            _ensure_cache(self)
            self.__cache__[name] = value

        def fdel(self):
            _ensure_cache(self)
            del self.__cache__[name]

        return property(wrapped, fset=fset, fdel=fdel)

    return property(wrapped)


def assert_valid_name(name: str):
    """Ensure a valid name of activity, state-machine or state.

    Args:
        name: name to analyse

    Raises:
        ValueError: name is invalid
    """

    if len(name) > MAX_NAME_LENGTH:
        raise ValueError("Name is too long: '%s'" % name)
    if any(c in name for c in INVALID_NAME_CHARACTERS):
        raise ValueError("Name contains invalid characters: '%s'" % name)


def collect_paginated(
        fn: T.Callable[..., T.Dict[str, JSONable]],
        **kwargs
) -> T.Dict[str, JSONable]:
    """Call SFN API paginated endpoint.

    Calls ``fn`` until "nextToken" isn't in the return value, collating
    results. Uses recursion: if recursion limit is reached, increase
    ``maxResults`` if available, otherwise increase the maximum recursion
    limit using the ``sys`` package.

    Args:
        fn: SFN API function
        **kwargs: arguments to ``fn``

    Returns:
        combined results of paginated API calls
    """

    result = fn(**kwargs)
    if "nextToken" in result:
        kwargs["nextToken"] = result.pop("nextToken")
        r2 = collect_paginated(fn, **kwargs)
        [result[k].extend(v) for k, v in r2.items() if isinstance(v, list)]
    return result


def easy_repr(instance) -> str:
    """Use attributes to generate a string representation.

    Set class ``__repr__ = easy_repr``.

    Args:
        instance: object to get representation of

    Returns:
        object representation
    """

    sig = inspect.signature(type(instance))
    params = sig.parameters.values()

    # Can't yet process var-args
    has_var_pos = any(p.kind == p.VAR_POSITIONAL for p in params)
    has_var_kw = any(p.kind == p.VAR_KEYWORD for p in params)
    if has_var_pos or has_var_kw:
        raise RuntimeError("Can't use `easy_repr` with var-args yet")

    # Separate difference kinds of parameters
    params_pos = [p for p in params if p.kind == p.POSITIONAL_ONLY]
    params_any = [p for p in params if p.kind == p.POSITIONAL_OR_KEYWORD]
    params_kw = [p for p in params if p.kind == p.KEYWORD_ONLY]

    params_any_required = [p for p in params_any if p.default == p.empty]
    params_any_optional = [p for p in params_any if p.default != p.empty]
    params_unnamed = params_pos + params_any_required
    params_named = params_any_optional + params_kw

    arg_strs = []
    for param in params_unnamed:
        attr_val = getattr(instance, param.name)
        arg_str = repr(attr_val)
        if len(arg_str) > 80 and isinstance(attr_val, abc.Sized):
            arg_str = "len %d" % len(attr_val)
        arg_strs.append(arg_str)
    for param in params_named:
        attr_val = getattr(instance, param.name)
        if param.default != param.empty and attr_val == param.default:
            continue
        arg_str = repr(attr_val)
        if len(arg_str) > 80 and isinstance(attr_val, abc.Sized):
            arg_str = "len(%s)=%d" % (param.name, len(attr_val))
        else:
            arg_str = "%s=%s" % (param.name, arg_str)
        arg_strs.append(arg_str)

    args_str = ", ".join(arg_strs)
    type_name = type(instance).__name__
    return "%s(%s)" % (type_name, args_str)


class AWSSession:
    """AWS session, for preconfigure communication with AWS.

    Args:
        session: session to use
    """

    def __init__(self, session: boto3.Session = None):
        self.session = session or boto3.Session()

    def __str__(self):
        fmt = "<access key: %s, region: %s>"
        return fmt % (self.credentials.access_key, self.region)

    __repr__ = easy_repr

    @cached_property
    def credentials(self) -> credentials.Credentials:
        """AWS session credentials."""
        return self.session.get_credentials()

    @cached_property
    def sfn(self) -> botocore_client.BaseClient:
        """Step Functions client."""
        return self.session.client("stepfunctions")

    @cached_property
    def region(self) -> str:
        """Session AWS region."""
        return self.session.region_name

    @cached_property
    def account_id(self) -> str:
        """Session's account's account ID."""
        return self.session.client("sts").get_caller_identity()["Account"]



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/07/11

"""State-machine interfacing.

A state-machine defines the logic for a workflow of an application. It
is comprised of states (ie stages), and executions of which will run
the workflow over some given data.
"""

import json
import uuid
import datetime
import typing as T
import logging as lg

from . import _util
from . import execution as sfini_execution
from . import state as sfini_state

_logger = lg.getLogger(__name__)
_default = _util.DefaultParameter()


class StateMachine:
    """State machine structure for AWS Step Functions.

    Args:
        name: name of state-machine
        states: state-machine states
        start_state: name of start state
        comment: description of state-maching
        timeout: execution time-out (seconds)
        session: session to use for AWS communication
    """

    _execution_class = sfini_execution.Execution

    def __init__(
            self,
            name: str,
            states: T.Dict[str, sfini_state.State],
            start_state: str,
            comment: str = _default,
            timeout: int = _default,
            *,
            session: _util.AWSSession = None):
        self.name = name
        self.states = states
        self.start_state = start_state
        self.comment = comment
        self.timeout = timeout
        self.session = session or _util.AWSSession()

    def __str__(self):
        return "'%s' (%d states)" % (self.name, len(self.states))

    __repr__ = _util.easy_repr

    @_util.cached_property
    def arn(self) -> str:
        """State-machine generated ARN."""
        region = self.session.region
        account = self.session.account_id
        fmt = "arn:aws:states:%s:%s:stateMachine:%s"
        return fmt % (region, account, self.name)

    @_util.cached_property
    def default_role_arn(self) -> str:
        """``sfini``-generated state-machine IAM role ARN."""
        return "arn:aws:iam::%s:role/sfiniGenerated" % self.session.account_id

    def to_dict(self) -> T.Dict[str, _util.JSONable]:
        """Convert this state-machine to a definition dictionary.

        Returns:
            definition
        """

        _logger.debug("Converting '%s' to dictionary" % self)
        state_defns = {n: s.to_dict() for n, s in self.states.items()}
        defn = {"StartAt": self.start_state, "States": state_defns}
        if self.comment != _default:
            defn["Comment"] = self.comment
        if self.timeout != _default:
            defn["TimeoutSeconds"] = self.timeout
        return defn

    def is_registered(self) -> bool:
        """See if this state-machine is registered with AWS SFN.

        Returns:
            if this state-machine is registered
        """

        _logger.debug("Testing for registration of '%s' on SFN" % self)
        resp = _util.collect_paginated(self.session.sfn.list_state_machines)
        arns = {sm["stateMachineArn"] for sm in resp["stateMachines"]}
        return self.arn in arns

    def _sfn_create(self, role_arn: str):
        """Create this state-machine in AWS SFN.

        Args:
            role_arn: state-machine IAM role ARN
        """

        _logger.info("Creating '%s' on SFN" % self)
        resp = self.session.sfn.create_state_machine(
            name=self.name,
            definition=json.dumps(self.to_dict(), indent=4),
            roleArn=role_arn)
        assert resp["stateMachineArn"] == self.arn
        fmt = "State-machine '%s' registered with ARN '%s' at %s"
        _logger.info(fmt % (self, self.arn, resp["creationDate"]))

    def _sfn_update(self, role_arn: str = None):
        """Update this state-machine in AWS SFN.

        Args:
            role_arn: state-machine IAM role ARN to update to
        """

        _logger.info("Updating '%s' on SFN" % self)
        kwargs = {} if role_arn is None else {"roleArn": role_arn}
        resp = self.session.sfn.update_state_machine(
            definition=json.dumps(self.to_dict(), indent=4),
            stateMachineArn=self.arn,
            **kwargs)
        _logger.info("'%s' updated at %s" % (self, resp["updateDate"]))

    def register(self, role_arn: str = None, allow_update: bool = False):
        """Register state-machine with AWS SFN.

        Args:
            role_arn: state-machine IAM role ARN
            allow_update: allow overwriting of an existing state-machine with
                the same name
        """

        _util.assert_valid_name(self.name)
        if allow_update and self.is_registered():
            self._sfn_update(role_arn)
        else:
            self._sfn_create(role_arn or self.default_role_arn)

    def deregister(self):
        """Remove state-machine from AWS SFN."""
        _logger.info("Deleting state-machine '%s' from SFN" % self)
        self.session.sfn.delete_state_machine(stateMachineArn=self.arn)

    def start_execution(
            self,
            execution_input: _util.JSONable
    ) -> _execution_class:
        """Start an execution.

        Args:
            execution_input: input to first state in state-machine

        Returns:
            started execution
        """

        fmt = "Starting execution of '%s' with: %s"
        _logger.info(fmt % (self, execution_input))

        _now = datetime.datetime.now().strftime("%Y-%m-%dT%H-%M-%S")
        name = self.name + "_" + _now + "_" + str(uuid.uuid4())[:8]
        execution = self._execution_class(
            name,
            self.arn,
            execution_input,
            session=self.session)
        execution.start()
        return execution

    def _build_executions(
            self,
            items: T.List[T.Dict[str, _util.JSONable]]
    ) -> T.List[_execution_class]:
        """Build executions from response list-items.

        This state-machine is manually attached to the ``state_machine``
        attribute of the resultant executions here.

        Arguments:
            items: execution list-items

        Returns:
            constructed executions
        """

        executions = []
        for item in items:
            assert item["stateMachineArn"] == self.arn
            execution = self._execution_class.from_execution_list_item(
                item,
                session=self.session)
            execution.state_machine = self
            executions.append(execution)

            fmt = "Found execution '%s' with stop-date: %s"
            _logger.debug(fmt % (execution, item.get("stopDate")))
        return executions

    def list_executions(self, status: str = None) -> T.List[_execution_class]:
        """List all executions of this state-machine.

        This state-machine is manually attached to the ``state_machine``
        attribute of the resultant executions here.

        Args:
            status: only list executions with this status. Choose from
                'RUNNING', 'SUCCEEDED', 'FAILED', 'TIMED_OUT' or 'ABORTED'

        Returns:
            executions of this state-machine
        """

        _s = " with status '%s'" % status if status else ""
        _logger.info("Listing executions of '%s'" % self + _s)

        kwargs = {"stateMachineArn": self.arn}
        if status is not None:
            kwargs["statusFilter"] = status
        fn = self.session.sfn.list_executions
        resp = _util.collect_paginated(fn, **kwargs)
        return self._build_executions(resp["executions"])


def construct_state_machine(
        name: str,
        start_state: sfini_state.State,
        comment: str = _default,
        timeout: int = _default,
        *,
        session: _util.AWSSession = None
) -> StateMachine:
    """Construct a state-machine from the starting state.

    Make sure to construct the state-machine after all states have been
    defined: subsequent states will need to be added to the state-machine
    manually.

    Only states referenced by the provided first state (and their children)
    will be in the state-machine definition. Add states via an impossible
    choice rule to include them in the definition.

    Args:
        name: name of state-machine
        start_state: starting state of state-machine
        comment: description of state-maching
        timeout: execution time-out (seconds)
        session: session to use for AWS communication

    Returns:
        constructed state-machine
    """

    states = {}
    start_state.add_to(states)
    return StateMachine(
        name,
        states,
        start_state.name,
        comment=comment,
        timeout=timeout,
        session=session)



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/07/12

"""Activity interfacing.

Activities are separate from state-machines, and are used as
implementations of 'Task' states. Activities are registered separately.
"""

import inspect
import typing as T
import logging as lg
import functools as ft

from . import _util
from . import task_resource as sfini_task_resource

_logger = lg.getLogger(__name__)


class Activity(sfini_task_resource.TaskResource):
    """Activity execution.

    Note that activity names must be unique (within a region). It's
    recommended to put your code's title and version in the activity name.
    ``Activities`` makes this straight-forward.

    An activity is attached to state-machine tasks, and is called when that
    task is executed. A worker registers itself able to run some
    activities using their names.

    Args:
        name: name of activity
        heartbeat: seconds between heartbeat during activity running
        session: session to use for AWS communication
    """

    service = "activity"

    def __init__(self, name, heartbeat: int = 20, *, session=None):
        super().__init__(name, session=session)
        self.heartbeat = heartbeat

    def register(self):
        """Register activity with AWS SFN."""
        _util.assert_valid_name(self.name)
        resp = self.session.sfn.create_activity(name=self.name)
        assert resp["activityArn"] == self.arn
        fmt = "Activity '%s' registered with ARN '%s' at %s"
        _logger.info(fmt % (self, self.arn, resp["creationDate"]))

    def is_registered(self) -> bool:
        """See if this activity is registered with AWS SFN.

        Returns:
            if this activity is registered
        """

        _logger.debug("Testing for registration of '%s' on SFN" % self)
        resp = _util.collect_paginated(self.session.sfn.list_activities)
        arns = {sm["activityArn"] for sm in resp["activities"]}
        return self.arn in arns

    def deregister(self):
        """Remove activity from AWS SFN."""
        _logger.info("Deleting activity '%s' from SFN" % self)
        self.session.sfn.delete_activity(activityArn=self.arn)


class CallableActivity(Activity):
    """Activity execution defined by a callable.

    Note that activity names must be unique (within a region). It's
    recommended to put your application's name and version in the activity
    name. ``ActivityRegistration`` makes this straight-forward.

    An activity is attached to state-machine tasks, and is called when that
    task is executed. A worker registers itself able to run an activity
    using the registered activity name.

    Args:
        name: name of activity
        fn: function to run activity
        heartbeat: seconds between heartbeat during activity running
        session: session to use for AWS communication
    """

    def __init__(self, name, fn: T.Callable, heartbeat=20, *, session=None):
        super().__init__(name, heartbeat=heartbeat, session=session)
        self.fn = fn

    def __call__(self, task_input: _util.JSONable, *args, **kwargs):
        return self.fn(task_input, *args, **kwargs)

    @classmethod
    def decorate(
            cls,
            name: str,
            heartbeat: int = 20,
            *,
            session: _util.AWSSession = None
    ) -> T.Callable[[T.Callable], "CallableActivity"]:
        """Decorate a callable as an activity implementation.

        Args:
            name: name of activity
            heartbeat: seconds between heartbeat during activity running
            session: session to use for AWS communication
        """

        def wrap(fn: T.Callable):
            activity = cls(name, fn, heartbeat=heartbeat, session=session)
            return ft.update_wrapper(activity, fn)
        return wrap

    def call_with(self, task_input: _util.JSONable) -> _util.JSONable:
        """Call with task-input context.

        Args:
            task_input: task input

        Returns:
            function return-value
        """

        return self.fn(task_input)


class SmartCallableActivity(CallableActivity):
    """Activity execution defined by a callable, processing input.

    The arguments to ``fn`` are extracted from the input provided by AWS
    Step Functions.

    Note that activity names must be unique (within a region). It's
    recommended to put your application's name and version in the activity
    name. ``ActivityRegistration`` makes this straight-forward.

    An activity is attached to state-machine tasks, and is called when that
    task is executed. A worker registers itself able to run an activity
    using the registered activity name.

    Args:
        name: name of activity
        fn: function to run activity
        heartbeat: seconds between heartbeat during activity running
        session: session to use for AWS communication

    Attributes:
        sig: function signature
    """

    def __init__(self, name, fn: T.Callable, heartbeat=20, *, session=None):
        super().__init__(name, fn, heartbeat=heartbeat, session=session)
        self.sig: inspect.Signature = inspect.Signature.from_callable(fn)

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

    def _get_input_from(
            self,
            task_input: T.Dict[str, _util.JSONable]
    ) -> T.Dict[str, _util.JSONable]:
        """Parse task input for execution input.

        Does not perform input validation: ``fn(**kwargs)`` in
        ``call_with`` will do that anyway.

        Args:
            task_input: task input

        Returns:
            activity input
        """

        kinds = {n: p.kind for n, p in self.sig.parameters.items()}
        if any(k == inspect.Parameter.VAR_KEYWORD for k in kinds.values()):
            return task_input

        var_pos = inspect.Parameter.VAR_POSITIONAL
        kwargs = {}
        for name, arg_val in task_input.items():
            if kinds.get(name, var_pos) != var_pos:
                kwargs[name] = arg_val
        return kwargs

    def call_with(self, task_input: T.Dict[str, _util.JSONable]):
        kwargs = self._get_input_from(task_input)
        return self.fn(**kwargs)


class ActivityRegistration:
    """Activities registration.

    Provides convenience for grouping activities, generating activity
    names, bulk-registering activities, and activity function decoration.

    An activity is attached to state-machine tasks, and is called when that
    task is executed. A worker registers itself able to run an activity
    using the registered activity name.

    Args:
        prefix: prefix for activity names
        session: session to use for AWS communication

    Attributes:
        activities: registered activities

    Example:
        >>> activities = ActivityRegistration(prefix="foo")
        >>> @activities.activity(name="MyActivity")
        >>> def fn(data):
        ...     print("hi")
        >>> print(fn.name)
        fooMyActivity
    """

    _activity_class = CallableActivity
    _smart_activity_class = SmartCallableActivity

    def __init__(self, prefix: str = "", *, session: _util.AWSSession = None):
        self.prefix = prefix
        self.session = session or _util.AWSSession()
        self.activities: T.Dict[str, Activity] = {}

    def __str__(self):
        return "'%s' activities" % self.prefix

    __repr__ = _util.easy_repr

    def add_activity(self, activity: Activity):
        """Add an activity to the group.

        Args:
            activity: activity to add

        Raises:
            ValueError: if activity name already in-use in group
        """

        if activity.name in self.activities:
            raise ValueError("Activity '%s' already in group" % activity.name)
        self.activities[activity.name] = activity

    def _activity(
            self,
            activity_cls: T.Type[CallableActivity],
            name: str = None,
            heartbeat: int = 20
    ) -> T.Callable[[T.Callable], CallableActivity]:
        """Activity function decorator.

        Args:
            activity_cls: activity class
            name: name of activity, default: function name
            heartbeat: seconds between heartbeat during activity running
        """

        def wrap(fn):
            suff = fn.__name__ if name is None else name
            activity = activity_cls.decorate(
                self.prefix + suff,
                heartbeat=heartbeat,
                session=self.session)(fn)
            self.add_activity(activity)
            return activity
        return wrap

    def activity(
            self,
            name: str = None,
            heartbeat: int = 20
    ) -> T.Callable[[T.Callable], CallableActivity]:
        """Activity function decorator.

        The decorated function will be passed one argument: the input to
        the task state that executes the activity.

        Args:
            name: name of activity, default: function name
            heartbeat: seconds between heartbeat during activity running
        """

        return self._activity(
            self._activity_class,
            name=name,
            heartbeat=heartbeat)

    def smart_activity(
            self,
            name: str = None,
            heartbeat: int = 20
    ) -> T.Callable[[T.Callable], SmartCallableActivity]:
        """Smart activity function decorator.

        The decorated function will be passed values to its parameters
        from the input to the task state that executes the activity.

        Args:
            name: name of activity, default: function name
            heartbeat: seconds between heartbeat during activity running
        """

        return self._activity(
            self._smart_activity_class,
            name=name,
            heartbeat=heartbeat)

    def register(self):
        """Add registered activities to AWS SFN."""
        for activity in self.activities.values():
            activity.register()

    def _list_activities(self) -> T.List[T.Tuple[str, str, str]]:
        """List activities in SFN."""
        resp = _util.collect_paginated(self.session.sfn.list_activities)
        acts = []
        for act in resp["activities"]:
            prefix = act["name"][:len(self.prefix)]
            if prefix != self.prefix and act["name"] not in self.activities:
                continue
            acts.append((act["name"], act["activityArn"], act["creationDate"]))
        return acts

    def _deregister_activities(
            self,
            activity_items: T.Sequence[T.Tuple[str, str, str]]):
        """Deregister activities."""
        _logger.info("Deregistering %d activities" % len(activity_items))
        for act in activity_items:
            _logger.debug("Deregistering '%s'" % act[0])
            self.session.sfn.delete_activity(activityArn=act[1])

    def deregister(self):
        """Remove activities in AWS SFN."""
        acts = self._list_activities()
        self._deregister_activities(acts)



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/08/09

"""State-machine execution history events.

Use ``sfini.execution.Execution.format_history`` for nice history
printing.
"""

import json
import typing as T
import logging as lg

from .. import _util

_logger = lg.getLogger(__name__)
_default = _util.DefaultParameter()
_type_keys = {
    "ActivityFailed": "activityFailedEventDetails",
    "ActivityScheduleFailed": "activityScheduleFailedEventDetails",
    "ActivityScheduled": "activityScheduledEventDetails",
    "ActivityStarted": "activityStartedEventDetails",
    "ActivitySucceeded": "activitySucceededEventDetails",
    "ActivityTimedOut": "activityTimedOutEventDetails",
    "ChoiceStateEntered": "stateEnteredEventDetails",
    "ChoiceStateExited": "stateExitedEventDetails",
    "ExecutionFailed": "executionFailedEventDetails",
    "ExecutionStarted": "executionStartedEventDetails",
    "ExecutionSucceeded": "executionSucceededEventDetails",
    "ExecutionAborted": "executionAbortedEventDetails",
    "ExecutionTimedOut": "executionTimedOutEventDetails",
    "FailStateEntered": "stateEnteredEventDetails",
    "LambdaFunctionFailed": "lambdaFunctionFailedEventDetails",
    "LambdaFunctionScheduleFailed": "lambdaFunctionScheduleFailedEventDetails",
    "LambdaFunctionScheduled": "lambdaFunctionScheduledEventDetails",
    "LambdaFunctionStartFailed": "lambdaFunctionStartFailedEventDetails",
    # "LambdaFunctionStarted": "lambdaFunctionStartedEventDetails",
    "LambdaFunctionSucceeded": "lambdaFunctionSucceededEventDetails",
    "LambdaFunctionTimedOut": "lambdaFunctionTimedOutEventDetails",
    "SucceedStateEntered": "stateEnteredEventDetails",
    "SucceedStateExited": "stateExitedEventDetails",
    # "TaskStateAborted": "stateAbortedEventDetails",
    "TaskStateEntered": "stateEnteredEventDetails",
    "TaskStateExited": "stateExitedEventDetails",
    "PassStateEntered": "stateEnteredEventDetails",
    "PassStateExited": "stateExitedEventDetails",
    # "ParallelStateAborted": "stateAbortedEventDetails",
    "ParallelStateEntered": "stateEnteredEventDetails",
    "ParallelStateExited": "stateExitedEventDetails",
    "ParallelStateFailed": "stateFailedEventDetails",
    # "ParallelStateStarted": "stateStartedEventDetails",
    # "ParallelStateSucceeded": "stateSucceededEventDetails",
    # "WaitStateAborted": "stateAbortedEventDetails",
    "WaitStateEntered": "stateEnteredEventDetails",
    "WaitStateExited": "stateExitedEventDetails"}


class Event:
    """An execution history event.

    Args:
        timestamp (datetime.datetime): event time-stamp
        event_type: type of event
        event_id: identifying index of event
        previous_event_id: identifying index of causal event
    """

    def __init__(
            self,
            timestamp,
            event_type: str,
            event_id: int,
            previous_event_id: int = None):
        self.timestamp = timestamp
        self.event_type = event_type
        self.event_id = event_id
        self.previous_event_id = previous_event_id

    def __str__(self):
        fmt = "%s [%s] @ %s"
        return fmt % (self.event_type, self.event_id, self.timestamp)

    __repr__ = _util.easy_repr

    @staticmethod
    def _get_args(
            history_event: T.Dict[str, _util.JSONable]
    ) -> T.Tuple[tuple, T.Dict[str, T.Any], T.Dict[str, _util.JSONable]]:
        """Get initialisation arguments by parsing history event.

        Args:
            history_event: execution history event, provided by AWS API

        Returns:
            initialisation positional and keyword arguments, and event details
        """

        # _logger.debug("history_event: %s" % history_event)
        timestamp = history_event["timestamp"]
        event_type = history_event["type"]
        event_id = history_event["id"]
        args = (timestamp, event_type, event_id)
        kwargs = {}
        if "previousEventId" in history_event:
            kwargs["previous_event_id"] = history_event["previousEventId"]
        details = history_event.get(_type_keys.get(event_type), {})
        return args, kwargs, details

    @classmethod
    def from_history_event(
            cls,
            history_event: T.Dict[str, _util.JSONable]
    ) -> "Event":
        """Parse an history event.

        Args:
            history_event: execution history event date, provided by AWS API

        Returns:
            Event: constructed execution history event
        """

        args, kwargs, _ = cls._get_args(history_event)
        return cls(*args, **kwargs)

    @_util.cached_property
    def details_str(self) -> str:
        """Format the event details.

        Returns:
            str: event details, formatted as string
        """

        return ""


class Failed(Event):
    """An execution history failure event.

    Args:
        timestamp: event time-stamp
        event_type: type of event
        event_id: identifying index of event
        previous_event_id: identifying index of causal event
        error: error type
        cause: failure details
    """

    def __init__(
            self,
            timestamp,
            event_type,
            event_id,
            previous_event_id=None,
            error: str = None,
            cause: str = None):
        super().__init__(
            timestamp,
            event_type,
            event_id,
            previous_event_id=previous_event_id)
        self.error = error
        self.cause = cause

    @classmethod
    def _get_args(cls, history_event):
        args, kwargs, details = super()._get_args(history_event)
        if "error" in details:
            kwargs["error"] = details["error"]
        if "cause" in details:
            kwargs["cause"] = details["cause"]
        return args, kwargs, details

    @_util.cached_property
    def details_str(self):
        return "error: %s" % self.error


class LambdaFunctionScheduled(Event):
    """An execution history AWS Lambda task-schedule event.

    Args:
        timestamp: event time-stamp
        event_type: type of event
        event_id: identifying index of event
        resource: AWS Lambda function ARN
        previous_event_id: identifying index of causal event
        task_input: task input
        timeout: time-out (seconds) of task execution
    """

    def __init__(
            self,
            timestamp,
            event_type,
            event_id,
            resource: str,
            previous_event_id=None,
            task_input: _util.JSONable = _default,
            timeout: int = None):
        super().__init__(
            timestamp,
            event_type,
            event_id,
            previous_event_id=previous_event_id)
        self.resource = resource
        self.task_input = task_input
        self.timeout = timeout

    @classmethod
    def _get_args(cls, history_event):
        args, kwargs, details = super()._get_args(history_event)
        args += (details["resource"],)
        if "input" in details:
            kwargs["task_input"] = json.loads(details["input"])
        if "timeoutInSeconds" in details:
            kwargs["timeout"] = details["timeoutInSeconds"]
        return args, kwargs, details

    @_util.cached_property
    def details_str(self):
        return "resource: %s" % self.resource


class ActivityScheduled(LambdaFunctionScheduled):
    """An execution history activity task-schedule event.

    Args:
        timestamp: event time-stamp
        event_type: type of event
        event_id: identifying index of event
        resource: AWS Lambda function ARN
        previous_event_id: identifying index of causal event
        task_input: task input
        timeout: time-out (seconds) of task execution
        heartbeat: heartbeat time-out (seconds)
    """

    def __init__(
            self,
            timestamp,
            event_type,
            event_id,
            resource,
            previous_event_id=None,
            task_input=_default,
            timeout=None,
            heartbeat: int = None):
        super().__init__(
            timestamp,
            event_type,
            event_id,
            resource,
            previous_event_id=previous_event_id,
            task_input=task_input,
            timeout=timeout)
        self.heartbeat = heartbeat

    @classmethod
    def _get_args(cls, history_event):
        args, kwargs, details = super()._get_args(history_event)
        if "heartbeatInSeconds" in details:
            kwargs["heartbeat"] = details["heartbeatInSeconds"]
        return args, kwargs, details


class ActivityStarted(Event):
    """An execution history activity task-start event.

    Args:
        timestamp: event time-stamp
        event_type: type of event
        event_id: identifying index of event
        worker_name: name of activity worker executing activity task
        previous_event_id: identifying index of causal event
    """

    def __init__(
            self,
            timestamp,
            event_type,
            event_id,
            worker_name: str,
            previous_event_id=None):
        super().__init__(
            timestamp,
            event_type,
            event_id,
            previous_event_id=previous_event_id)
        self.worker_name = worker_name

    @classmethod
    def _get_args(cls, history_event):
        args, kwargs, details = super()._get_args(history_event)
        args += (details["workerName"],)
        return args, kwargs, details

    @_util.cached_property
    def details_str(self):
        return "worker: %s" % self.worker_name


class ObjectSucceeded(Event):
    """An execution history succeed event.

    Args:
        timestamp: event time-stamp
        event_type: type of event
        event_id: identifying index of event
        previous_event_id: identifying index of causal event
        output: output of state/execution
    """

    def __init__(
            self,
            timestamp,
            event_type,
            event_id,
            previous_event_id=None,
            output: _util.JSONable = _default):
        super().__init__(
            timestamp,
            event_type,
            event_id,
            previous_event_id=previous_event_id)
        self.output = output

    @classmethod
    def _get_args(cls, history_event):
        args, kwargs, details = super()._get_args(history_event)
        if "output" in details:
            kwargs["output"] = json.loads(details["output"])
        return args, kwargs, details


class ExecutionStarted(Event):
    """An execution history execution-start event.

    Args:
        timestamp: event time-stamp
        event_type: type of event
        event_id: identifying index of event
        previous_event_id: identifying index of causal event
        execution_input: execution input
        role_arn: execution AWS IAM role ARN
    """

    def __init__(
            self,
            timestamp,
            event_type,
            event_id,
            previous_event_id=None,
            execution_input: _util.JSONable = _default,
            role_arn: str = None):
        super().__init__(
            timestamp,
            event_type,
            event_id,
            previous_event_id=previous_event_id)
        self.execution_input = execution_input
        self.role_arn = role_arn

    @classmethod
    def _get_args(cls, history_event):
        args, kwargs, details = super()._get_args(history_event)
        if "input" in details:
            kwargs["execution_input"] = json.loads(details["input"])
        if "roleArn" in details:
            kwargs["role_arn"] = details["roleArn"]
        return args, kwargs, details


class StateEntered(Event):
    """An execution history state-enter event.

    Args:
        timestamp: event time-stamp
        event_type: type of event
        event_id: identifying index of event
        state_name: state name
        previous_event_id: identifying index of causal event
        state_input: state input
    """

    def __init__(
            self,
            timestamp,
            event_type,
            event_id,
            state_name: str,
            previous_event_id=None,
            state_input: _util.JSONable = _default):
        super().__init__(
            timestamp,
            event_type,
            event_id,
            previous_event_id=previous_event_id)
        self.state_name = state_name
        self.state_input = state_input

    @classmethod
    def _get_args(cls, history_event):
        args, kwargs, details = super()._get_args(history_event)
        args += (details["name"],)
        if "input" in details:
            kwargs["state_input"] = json.loads(details["input"])
        return args, kwargs, details

    @_util.cached_property
    def details_str(self):
        return "name: %s" % self.state_name


class StateExited(Event):
    """An execution history state-exit event.

    Args:
        timestamp: event time-stamp
        event_type: type of event
        event_id: identifying index of event
        state_name: state name
        previous_event_id: identifying index of causal event
        output: state output
    """

    def __init__(
            self,
            timestamp,
            event_type,
            event_id,
            state_name: str,
            previous_event_id=None,
            output: _util.JSONable = _default):
        super().__init__(
            timestamp,
            event_type,
            event_id,
            previous_event_id=previous_event_id)
        self.state_name = state_name
        self.output = output

    @classmethod
    def _get_args(cls, history_event):
        args, kwargs, details = super()._get_args(history_event)
        args += (details["name"],)
        if "output" in details:
            kwargs["output"] = json.loads(details["output"])
        return args, kwargs, details

    @_util.cached_property
    def details_str(self):
        return "name: %s" % self.state_name


_type_classes = {
    "ActivityFailed": Failed,
    "ActivityScheduleFailed": Failed,
    "ActivityScheduled": ActivityScheduled,
    "ActivityStarted": ActivityStarted,
    "ActivitySucceeded": ObjectSucceeded,
    "ActivityTimedOut": Failed,
    "ChoiceStateEntered": StateEntered,
    "ChoiceStateExited": StateExited,
    "ExecutionFailed": Failed,
    "ExecutionStarted": ExecutionStarted,
    "ExecutionSucceeded": ObjectSucceeded,
    "ExecutionAborted": Failed,
    "ExecutionTimedOut": Failed,
    "FailStateEntered": StateEntered,
    "LambdaFunctionFailed": Failed,
    "LambdaFunctionScheduleFailed": Failed,
    "LambdaFunctionScheduled": LambdaFunctionScheduled,
    "LambdaFunctionStartFailed": Failed,
    "LambdaFunctionStarted": Event,
    "LambdaFunctionSucceeded": ObjectSucceeded,
    "LambdaFunctionTimedOut": Failed,
    "SucceedStateEntered": StateEntered,
    "SucceedStateExited": StateExited,
    "TaskStateAborted": Event,
    "TaskStateEntered": StateEntered,
    "TaskStateExited": StateExited,
    "PassStateEntered": StateEntered,
    "PassStateExited": StateExited,
    "ParallelStateAborted": Event,
    "ParallelStateEntered": StateEntered,
    "ParallelStateExited": StateExited,
    "ParallelStateFailed": Failed,
    "ParallelStateStarted": Event,
    "ParallelStateSucceeded": Event,
    "WaitStateAborted": Event,
    "WaitStateEntered": StateEntered,
    "WaitStateExited": StateExited}


def parse_history(
        history_events: T.List[T.Dict[str, _util.JSONable]]
) -> T.List[Event]:
    """List the execution history.

    Args:
        history_events: history events as provided by AWS API

    Returns:
        history of execution events
    """

    events = []
    for history_event in history_events:
        eclass = _type_classes[history_event["type"]]
        event = eclass.from_history_event(history_event)
        events.append(event)
    return events



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/07/12

"""Activity task polling and execution.

You can provide you're own workers: the interface to the activities is
public. This module's worker implementation uses threading, and is
designed to be resource-managed outside of Python.
"""

import json
import uuid
import time
import socket
import threading
import traceback
import typing as T
import logging as lg

from botocore import exceptions as bc_exc

from . import _util

_logger = lg.getLogger(__name__)
_host_name = socket.getfqdn(socket.gethostname())


class WorkerCancel(KeyboardInterrupt):
    """Workflow execution interrupted by user."""
    def __init__(self, *args, **kwargs):
        msg = (
            "Activity execution cancelled by user. "
            "This could be due to a `KeyboardInterrupt` during execution, "
            "or the worker was killed during task polling.")
        super().__init__(msg, *args, **kwargs)


class TaskExecution:
    """Execute a task, providing heartbeats and catching failures.

    Args:
        activity (sfini.activity.CallableActivity): activity to execute
            task of
        task_token: task token for execution identification
        task_input: task input
        session: session to use for AWS communication
    """

    def __init__(
            self,
            activity,
            task_token: str,
            task_input: _util.JSONable,
            *,
            session: _util.AWSSession = None):
        self.activity = activity
        self.task_token = task_token
        self.task_input = task_input
        self.session = session or _util.AWSSession()

        self._heartbeat_thread = threading.Thread(target=self._heartbeat)
        self._request_stop = False

    def __str__(self):
        return "%s - %s" % (self.activity.name, self.task_token)

    __repr__ = _util.easy_repr

    def _send(self, send_fn: T.Callable, **kw):
        """Send execution update to SFN."""
        if self._request_stop:
            _logger.warning("Skipping sending update as we've already quit")
            return
        return send_fn(taskToken=self.task_token, **kw)

    def _report_exception(self, exc: BaseException):
        """Report failure."""
        _logger.info("Reporting task failure for '%s'" % self, exc_info=exc)
        tb = traceback.format_exception(type(exc), exc, exc.__traceback__)
        self._send(
            self.session.sfn.send_task_failure,
            error=type(exc).__name__,
            cause="".join(tb))
        self._request_stop = True

    def report_cancelled(self):
        """Cancel a task execution: stop interaction with SFN."""
        fmt = "Reporting task failure for '%s' due to cancellation"
        _logger.info(fmt % self)
        self._send(
            self.session.sfn.send_task_failure,
            error=WorkerCancel.__name__,
            cause=str(WorkerCancel()))
        self._request_stop = True

    def _report_success(self, res: _util.JSONable):
        """Report success."""
        fmt = "Reporting task success for '%s' with output: %s"
        _logger.debug(fmt % (self, res))
        self._send(self.session.sfn.send_task_success, output=json.dumps(res))
        self._request_stop = True

    def _send_heartbeat(self):
        """Send a heartbeat."""
        _logger.debug("Sending heartbeat for '%s'" % self)
        try:
            self._send(self.session.sfn.send_task_heartbeat)
        except bc_exc.ClientError as e:
            self._request_stop = True
            if e.response["Error"]["Code"] != "TaskTimedOut":
                raise
            _logger.error("Task execution '%s' timed-out" % self)

    def _heartbeat(self):
        """Run heartbeat sending."""
        heartbeat = self.activity.heartbeat
        while True:
            t = time.time()
            if self._request_stop:
                break
            self._send_heartbeat()
            time.sleep(heartbeat - (time.time() - t))

    def run(self):
        """Run task."""
        self._heartbeat_thread.start()
        t = time.time()

        try:
            res = self.activity.call_with(self.task_input)
        except KeyboardInterrupt:
            self.report_cancelled()
            return
        except Exception as e:
            self._report_exception(e)
            return

        fmt = "Task '%s' completed in %.6f seconds"
        _logger.debug(fmt % (self, time.time() - t))
        self._report_success(res)


class Worker:
    """Worker to poll for activity task executions.

    Args:
        activity (sfini.activity.CallableActivity): activity to poll and
            run executions of
        name: name of worker, used for identification, default: a
            combination of UUID and host's FQDN
        session: session to use for AWS communication
    """

    _task_execution_class = TaskExecution

    def __init__(
            self,
            activity,
            name: str = None,
            *,
            session: _util.AWSSession = None):
        self.activity = activity
        self.name = name or "%s-%s" % (_host_name, str(str(uuid.uuid4()))[:8])
        self.session = session or _util.AWSSession()

        self._poller = threading.Thread(target=self._worker)
        self._request_finish = False
        self._exc = None

    def __str__(self):
        return "%s [%s]" % (self.name, self.activity.name)

    __repr__ = _util.easy_repr

    def _execute_on(self, task_input: _util.JSONable, task_token: str):
        """Execute the provided task.

        Args:
            task_input: activity task execution input
            task_token: task execution identifier
        """

        _logger.debug("Got task input: %s" % task_input)

        execution = self._task_execution_class(
            self.activity,
            task_token,
            task_input,
            session=self.session)
        if self._request_finish:
            execution.report_cancelled()
        else:
            execution.run()

    def _poll_and_execute(self):
        """Poll for tasks to execute, then execute any found."""
        while not self._request_finish:
            fmt = "Polling for activity '%s' executions"
            _logger.debug(fmt % self.activity)
            resp = self.session.sfn.get_activity_task(
                activityArn=self.activity.arn,
                workerName=self.name)
            if resp.get("taskToken", None) is not None:
                input_ = json.loads(resp["input"])
                self._execute_on(input_, resp["taskToken"])

    def _worker(self):
        """Run polling, catching exceptins."""
        try:
            self._poll_and_execute()
        except (Exception, KeyboardInterrupt) as e:
            _logger.warning("Polling/execution failed", exc_info=e)
            self._exc = e  # send exception to main thread
            self._request_finish = True

    def start(self):
        """Start polling."""
        from . import activity
        if not isinstance(self.activity, activity.CallableActivity):
            raise TypeError("Activity '%s' cannot be executed" % self.activity)
        _util.assert_valid_name(self.name)
        _logger.info("Worker '%s': waiting on final poll to finish" % self)
        self._poller.start()

    def join(self):
        """Block until polling exit."""
        try:
            self._poller.join()
        except KeyboardInterrupt:
            _logger.info("Quitting polling due to KeyboardInterrupt")
            self._request_finish = True
            return
        except Exception:
            self._request_finish = True
            raise
        if self._exc is not None:
            raise self._exc

    def end(self):
        """End polling."""
        _logger.info("Worker '%s': waiting on final poll to finish" % self)
        self._request_finish = True

    def run(self):
        """Run worker to poll for and execute specified tasks."""
        self.start()
        self.join()



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/07/11

"""SFN choice rules.

These rules are used in the 'Choice' state of a state-machine, and
allow for conditional branching in the state-machine. There are two
types of choice rule: comparisons and logical operations.
"""

import datetime
import typing as T
import logging as lg

from .. import _util

_logger = lg.getLogger(__name__)


class ChoiceRule:
    """A choice case for the 'Choice' state.

    Args:
        next_state (sfini.state.State): state to execute on success
    """

    _final = False

    def __init__(self, next_state=None):
        self.next_state = next_state

    __repr__ = _util.easy_repr

    def _get_comparison(self) -> _util.JSONable:
        """Get this rule's comparison.

        Returns:
            comparison definition
        """

        raise NotImplementedError

    def to_dict(self) -> T.Dict[str, _util.JSONable]:
        """Convert this rule to a definition dictionary.

        Returns:
            definition
        """

        op_name = type(self).__name__
        if not self._final:
            raise RuntimeError("'%s' is not a valid choice rule" % op_name)
        comp = self._get_comparison()
        defn = {op_name: comp}
        if self.next_state:
            defn["Next"] = self.next_state.name
        return defn


class Comparison(ChoiceRule):
    """Compare variable value.

    Args:
        variable_path: path of variable to compare
        comparison_value: value to compare against
        next_state: state to execute on success
    """

    _expected_value_type = None

    def __init__(
            self,
            variable_path: str,
            comparison_value,
            next_state=None):
        super().__init__(next_state)
        self.variable_path = variable_path
        self.comparison_value = comparison_value

    def __str__(self):
        return "'%s' %s %s%s" % (
            self.variable_path,
            type(self).__name__,
            self.comparison_value,
            "" if self.next_state is None else (" -> %s" % self.next_state))

    def _get_comparison(self):
        if not isinstance(self.comparison_value, self._expected_value_type):
            fmt = "Comparison value must be type `%s`: %s"
            exp_type_name = self._expected_value_type.__name__
            raise TypeError(fmt % (self.comparison_value, exp_type_name))
        return self.comparison_value

    def to_dict(self):
        defn = super().to_dict()
        defn["Variable"] = self.variable_path
        return defn


class BooleanEquals(Comparison):
    _final = True
    _expected_value_type = bool


class NumericEquals(Comparison):
    _final = True
    _expected_value_type = (float, int)


class NumericGreaterThan(Comparison):
    _final = True
    _expected_value_type = (float, int)


class NumericGreaterThanEquals(Comparison):
    _final = True
    _expected_value_type = (float, int)


class NumericLessThan(Comparison):
    _final = True
    _expected_value_type = (float, int)


class NumericLessThanEquals(Comparison):
    _final = True
    _expected_value_type = (float, int)


class StringEquals(Comparison):
    _final = True
    _expected_value_type = str


class StringGreaterThan(Comparison):
    _final = True
    _expected_value_type = str


class StringGreaterThanEquals(Comparison):
    _final = True
    _expected_value_type = str


class StringLessThan(Comparison):
    _final = True
    _expected_value_type = str


class StringLessThanEquals(Comparison):
    _final = True
    _expected_value_type = str


class _TimestampRule(Comparison):
    _expected_value_type = datetime.datetime

    def _get_comparison(self) -> str:
        dt = super()._get_comparison()
        if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
            raise ValueError("Comparison time must be aware")
        return dt.isoformat("T")


class TimestampEquals(_TimestampRule):
    _final = True


class TimestampGreaterThan(_TimestampRule):
    _final = True


class TimestampGreaterThanEquals(_TimestampRule):
    _final = True


class TimestampLessThan(_TimestampRule):
    _final = True


class TimestampLessThanEquals(_TimestampRule):
    _final = True


class Logical(ChoiceRule):
    @staticmethod
    def _get_rule_defn(choice_rule: ChoiceRule) -> T.Dict[str, _util.JSONable]:
        """Get choice rule definition.

        Arguments:
            choice_rule: choice rule to process

        Returns:
            choice rule definition
        """

        if choice_rule.next_state is not None:
            msg = "Only top-level choice rules can have next state"
            raise RuntimeError(msg)
        return choice_rule.to_dict()


class _NonUnary(Logical):
    """Logical operation on choice rules.

    Args:
        choice_rules: choice rules to operate on
        next_state: state to execute on success
    """

    def __init__(self, choice_rules: T.List[ChoiceRule], next_state=None):
        super().__init__(next_state=next_state)
        self.choice_rules = choice_rules

    def __str__(self):
        _t = " %s " % type(self).__name__
        _s = _t.join("(%s)" % r for r in self.choice_rules)
        _n = "" if self.next_state is None else (" -> %s" % self.next_state)
        return _s + _n

    def _get_comparison(self) -> T.List[T.Dict[str, _util.JSONable]]:
        if not self.choice_rules:
            msg = "Must provide at least one choice-rule to logical choice"
            raise ValueError(msg)
        return [self._get_rule_defn(r) for r in self.choice_rules]


class And(_NonUnary):
    _final = True


class Or(_NonUnary):
    _final = True


class Not(Logical):
    """Logical 'not' operation on a choice rule.

    Args:
        choice_rule: choice rule to operate on
        next_state: state to execute on success
    """

    _final = True

    def __init__(self, choice_rule: ChoiceRule, next_state=None):
        super().__init__(next_state=next_state)
        self.choice_rule = choice_rule

    def __str__(self):
        _n = "" if self.next_state is None else (" -> %s" % self.next_state)
        _f = (type(self).__name__, self.choice_rule, _n)
        return "%s %s%s" % _f

    def _get_comparison(self) -> T.Dict[str, _util.JSONable]:
        return self._get_rule_defn(self.choice_rule)



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/08/06

"""SFN service CLI helper.

Use in your ``__main__`` module to provide a CLI to your service.
"""

import sys
import json
import pathlib
import argparse
import logging as lg

from . import _util
from . import worker as sfini_worker


class CLI:
    """``sfini`` command-line interface.

    Args:
        state_machine (sfini.StateMachine): state-machine interact with
        activities (sfini.ActivityRegistration): activities to poll for
        role_arn: AWS ARN for state-machine IAM role
        version: version to display, default: no version display
        prog : program name displayed in program help,
            default: ``sys.argv[0]``
    """

    _worker_class = sfini_worker.Worker
    _parser_class = argparse.ArgumentParser

    def __init__(
            self,
            state_machine=None,
            activities=None,
            role_arn: str = None,
            version: str = None,
            prog: str = None):
        self.state_machine = state_machine
        self.activities = activities
        self.role_arn = role_arn
        self.version = version
        self.prog = prog
        assert state_machine or activities

    __repr__ = _util.easy_repr

    def _build_parser(self) -> argparse.ArgumentParser:
        """Build argument parser.

        Returns:
            configured command-line argument parser
        """

        d = None
        if self.state_machine and self.activities:
            fmt = "Control %s and %s"
            d = fmt % (self.state_machine, self.activities)
        elif self.state_machine:
            d = "Control %s" % self.state_machine
        elif self.activities:
            d = "Control %s" % self.activities
        parser = self._parser_class(description=d, prog=self.prog)
        if self.version:
            parser.add_argument(
                "-V",
                "--version",
                action="version",
                version=self.version)
        parser.add_argument(
            "-v",
            "--verbose",
            default=0,
            action="count",
            help="increase verbosity")
        parser.add_argument(
            "-q",
            "--quiet",
            default=0,
            action="count",
            help="decrease verbosity")
        subparsers = parser.add_subparsers(
            metavar="COMMAND",
            # help="description",
            dest="command")
        subparsers.required = True  # Python 3.6 compatibility

        sma_str = {
            (True, True): "state-machine and/or activities",
            (True, False): "state-machine",
            (False, True): "activities"}
        sma_str = sma_str[bool(self.state_machine), bool(self.activities)]

        register_parser = subparsers.add_parser(
            "register",
            help="register %s with SFN" % sma_str,
            description="register %s with SFN" % sma_str)
        if self.state_machine:
            register_parser.add_argument(
                "-u",
                "--allow-update",
                action="store_true",
                help="allow updating of existing state-machine")
        if self.state_machine and self.activities:
            _g = register_parser.add_mutually_exclusive_group()
            _g.add_argument(
                "-s",
                "--state-machine-only",
                action="store_true",
                help="only register (or update) state-machine")
            _g.add_argument(
                "-a",
                "--activities-only",
                action="store_true",
                help="only register activities")

        deregister_parser = subparsers.add_parser(
            "deregister",
            help="deregister %s from SFN" % sma_str,
            description="deregister %s from SFN" % sma_str)
        if self.state_machine and self.activities:
            _g = deregister_parser.add_mutually_exclusive_group()
            _g.add_argument(
                "-s",
                "--state-machine-only",
                action="store_true",
                help="only deregister state-machine")
            _g.add_argument(
                "-a",
                "--activities-only",
                action="store_true",
                help="only deregister activities")

        if self.state_machine:
            start_parser = subparsers.add_parser(
                "start",
                help="start state-machine execution")
            start_parser.add_argument(
                "input_json",
                metavar="PATH",
                help="execution input JSON, use '-' for STDIN")
            start_parser.add_argument(
                "-w",
                "--wait",
                action="store_true",
                help="wait for execution to finish, and print output")

        if self.activities:
            worker_parser = subparsers.add_parser(
                "worker",
                help="run an activity worker",
                description="run an activity worker")
            worker_parser.add_argument(
                "activity_name",
                choices=self.activities.activities,
                metavar="NAME",
                help="name of activity to poll, choose from: %(choices)s)")

        if self.state_machine:
            executions_parser = subparsers.add_parser(
                "executions",
                help="list executions",
                description="list executions")
            choices = "RUNNING", "SUCCEEDED", "FAILED", "TIMED_OUT", "ABORTED"
            executions_parser.add_argument(
                "-s",
                "--status",
                default=None,
                metavar="STATUS",
                choices=choices,
                help="only list executions with this status")

        return parser

    def _register(self, args: argparse.Namespace):
        """Register state-machine and/or activities.

        Args:
            args: parsed command-line arguments
        """

        if self.state_machine and self.activities:
            if args.state_machine_only:
                self.state_machine.register(
                    self.role_arn,
                    allow_update=args.allow_update)
            elif args.activities_only:
                self.activities.register()
            else:
                self.state_machine.register(
                    self.role_arn,
                    allow_update=args.allow_update)
                self.activities.register()
        elif self.state_machine:
            if not args.activities_only:
                self.state_machine.register(
                    self.role_arn,
                    allow_update=args.allow_update)
        elif self.activities:
            if not args.state_machine_only:
                self.activities.register()

    def _deregister(self, args: argparse.Namespace):
        """Deregister state-machine and/or activities.

        Args:
            args: parsed command-line arguments
        """

        if self.state_machine and self.activities:
            if args.state_machine_only:
                self.state_machine.deregister()
            elif args.activities_only:
                self.activities.deregister()
            else:
                self.state_machine.deregister()
                self.activities.deregister()
        elif self.state_machine:
            if not args.activities_only:
                self.state_machine.deregister()
        elif self.activities:
            if not args.state_machine_only:
                self.activities.deregister()

    def _start(self, args: argparse.Namespace):
        """Start a state-machine execution.

        Args:
            args: parsed command-line arguments
        """

        if args.input_json == "-":
            execution_input_str = sys.stdin.read()
        else:
            input_json = pathlib.Path(args.input_json)
            execution_input_str = input_json.read_text()
        execution_input = json.loads(execution_input_str)
        execution = self.state_machine.start_execution(execution_input)
        if args.wait:
            execution.wait()
            print(execution.output)

    def _worker(self, args: argparse.Namespace):
        """Run an activity worker.

        Args:
            args: parsed command-line arguments
        """

        activity = self.activities.activities[args.activity_name]
        workers = self._worker_class(activity)
        workers.run()

    def _executions(self, args: argparse.Namespace):
        """List state-machine executions.

        Args:
            args: parsed command-line arguments
        """

        execs = self.state_machine.list_executions(status=args.status)
        for execution in execs:
            print(execution.format_history())

    def _delegate(self, args: argparse.Namespace):
        """Execute command.

        Args:
            args: parsed command-line arguments
        """

        _lvl = max(lg.WARNING - 10 * (args.verbose - args.quiet), lg.DEBUG)
        _util.setup_logging(level=_lvl)

        command = {
            "register": self._register,
            "deregister": self._deregister,
            "start": self._start,
            "worker": self._worker,
            "executions": self._executions}
        command[args.command](args)

    def parse_args(self):
        """Parse command-line arguments and run CLI."""
        _util.setup_logging()
        parser = self._build_parser()
        args = parser.parse_args()
        self._delegate(args)



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/07/11

"""State-machine execution."""

import time
import json
import datetime
import typing as T
import logging as lg

from .. import _util
from . import history

_logger = lg.getLogger(__name__)
_default = _util.DefaultParameter()


class Execution:
    """A state-machine execution.

    Args:
        name: name of execution
        state_machine_arn: executed state-machine ARN
        execution_input: execution input (must be JSON-serialisable)
        arn: execution ARN (if known: provided when execution is posted to
            AWS SFN)
        session: session to use for AWS communication
    """

    _wait_sleep_time = 3.0
    _not_provided = object()

    def __init__(
            self,
            name: str,
            state_machine_arn: str,
            execution_input: _util.JSONable = _default,
            arn: str = None,
            *,
            session: _util.AWSSession = None):
        self.name = name
        self.state_machine_arn = state_machine_arn
        self.execution_input = execution_input
        self.arn = arn
        self.session = session or _util.AWSSession()

        self._status = None
        self._start_date = None
        self._stop_date = None
        self._output = _default

    def __str__(self):
        status_str = (" [%s]" % self._status) if self._status else ""
        return "%s%s" % (self.name, status_str)

    __repr__ = _util.easy_repr

    @classmethod
    def from_arn(
            cls,
            arn: str,
            *,
            session: _util.AWSSession = None
    ) -> "Execution":
        """Construct an ``Execution`` from an existing execution.

        Queries AWS Step Functions for the execution with the given ARN

        Args:
            arn: existing execution ARN
            session: session to use for AWS communication

        Returns:
            described execution
        """

        session = session or _util.AWSSession()
        resp = session.sfn.describe_execution(executionArn=arn)
        assert resp["executionArn"] == arn
        execution_input = _default
        if "input" in resp:
            execution_input = json.loads(resp["input"])
        self = cls(
            resp["name"],
            resp["stateMachineArn"],
            execution_input=execution_input,
            arn=arn,
            session=session)
        self._status = resp["status"]
        self._start_date = resp["startDate"]
        self._stop_date = resp.get("stopDate")
        if "output" in resp:
            self._output = json.loads(resp["output"])
        return self

    @classmethod
    def from_execution_list_item(
            cls,
            item: T.Dict[str, _util.JSONable],
            *,
            session: _util.AWSSession = None
    ) -> "Execution":
        """Construct an ``Execution`` from a response list-item.

        Args:
            item: execution list item
            session: session to use for AWS communication

        Returns:
            described execution
        """

        self = cls(
            item["name"],
            item["stateMachineArn"],
            cls._not_provided,
            arn=item["executionArn"],
            session=session)
        self._status = item["status"]
        self._start_date = item["startDate"]
        self._stop_date = item.get("stopDate")
        return self

    @property
    def status(self) -> str:
        """Execution status."""
        if self._status in (None, "RUNNING"):
            self._update()
        return self._status

    @property
    def start_date(self) -> datetime.datetime:
        """Execution start time."""
        if self._start_date is None:
            self._update()
        return self._start_date

    @property
    def stop_date(self) -> datetime.datetime:
        """Execution stop time.

        Raises:
            RuntimeError: if execution is not yet finished
        """

        if self._stop_date is None:
            self._update()
            self._raise_unfinished()
        return self._stop_date

    @property
    def output(self) -> _util.JSONable:
        """Output of execution.

        Raises:
            RuntimeError: if execution is not yet finished, or execution
                failed
        """

        if self._output == _default:
            self._update()
            self._raise_unfinished()
            self._raise_on_failure()
        return self._output

    def _update(self):
        """Update execution information from AWS.

        Raises:
            RuntimeError: if execution ARN is not known (must be provided)
        """

        status_known = self._status not in (None, "RUNNING")
        input_known = self.execution_input != self._not_provided
        if status_known and input_known:
            _logger.debug("Execution finished: update is unnecessary")
            return
        self._raise_no_arn()
        resp = self.session.sfn.describe_execution(executionArn=self.arn)
        assert resp["executionArn"] == self.arn
        self._status = resp["status"]
        self._start_date = resp["startDate"]
        self._stop_date = resp.get("stopDate")
        if "input" in resp:
            input_ = json.loads(resp["input"])
            if self.execution_input == self._not_provided:
                self.execution_input = input_
            else:
                assert self.execution_input == input_
        if "output" in resp:
            self._output = json.loads(resp["output"])

    def _raise_on_failure(self):
        """Raise ``RuntimeError`` on execution failure."""
        failed = self._status not in ("RUNNING", "SUCCEEDED")
        if failed:
            raise RuntimeError("Execution '%s' %s" % (self, self._status))

    def _raise_unfinished(self):
        """Raise ``RuntimeError`` when requiring execution to be finished."""
        if self._status == "RUNNING":
            raise RuntimeError("Execution '%s' not yet finished" % self)

    def _raise_no_arn(self):
        """Raise ``RuntimeError`` when ARN is not provided."""
        if self.arn is None:
            raise RuntimeError("Execution '%s' ARN is unknown" % self)

    def start(self):
        """Start this state-machine execution.

        Sets the ``arn`` attribute.
        """

        _util.assert_valid_name(self.name)
        input_ = self.execution_input
        if input_ == _default:
            input_ = {}
        resp = self.session.sfn.start_execution(
            stateMachineArn=self.state_machine_arn,
            name=self.name,
            input=json.dumps(input_))
        self.arn = resp["executionArn"]
        self._status = "RUNNING"
        self._start_date = resp["startDate"]

    def wait(self, raise_on_failure: bool = True, timeout: float = None):
        """Wait for execution to finish.

        Args:
            raise_on_failure: raise error when execution fails
            timeout: time to wait for execution to finish (seconds),
                default: no time-out

        Raises:
            RuntimeError: if execution fails, or if time-out is reached
                before execution finishes
        """

        t = time.time()
        while True:
            self._update()
            if self._status != "RUNNING":
                break
            if timeout is not None and time.time() - t > timeout:
                raise RuntimeError("Time-out waiting on execution '%s'" % self)
            time.sleep(self._wait_sleep_time)
        if raise_on_failure:
            self._raise_on_failure()

    def stop(self, error_code: str = _default, details: str = _default):
        """Stop an existing execution.

        Args:
            error_code: stop reason identification
            details: stop reason

        Raises:
            RuntimeError: if execution is already finished
        """

        kw = {}
        if error_code != _default:
            kw["error"] = error_code
        if details != _default:
            kw["cause"] = details
        self._raise_no_arn()
        resp = self.session.sfn.stop_execution(executionArn=self.arn, **kw)
        self._stop_date = resp["stopDate"]
        _logger.info("Execution stopped on %s" % resp["stopDate"])

    def get_history(self) -> T.List[history.Event]:
        """List the execution history.

        Returns:
            history of execution events
        """

        self._raise_no_arn()
        resp = _util.collect_paginated(
            self.session.sfn.get_execution_history,
            executionArn=self.arn)
        return history.parse_history(resp["events"])

    def format_history(self) -> str:
        """Format the execution history for printing.

        Returns:
            history formatted
        """

        events = self.get_history()
        lines = []
        for event in events:
            ds = event.details_str
            line = ("%s:\n  %s" % (event, ds)) if ds else str(event)
            lines.append(line)
        self._update()
        if self._output != _default:
            line = "Output: %s" % json.dumps(self._output)
            lines.append(line)
        return "\n".join(lines)



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2019/05/13

"""State-machine execution interfacing.

Executions track state-machine execution history, input, status and (if
available) output. You can wait on it to finish, and iterate over its
history.
"""

__all__ = ["Execution", "history"]

from ._execution import Execution
from . import history



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/07/11

"""Install ``sfini``."""

import pathlib
import setuptools

parent = pathlib.Path(__file__).parent
long_description = (parent / "README.md").read_text()
version = (parent / "VERSION").read_text().strip()
classifiers = [
    "Environment :: Console",
    "Intended Audience :: Developers",
    "Programming Language :: Python :: 3 :: Only",
    "Natural Language :: English",
    "Operating System :: POSIX :: Linux",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: MacOS",
    "License :: OSI Approved :: MIT License"]
extras_require = {
        "dev": [
            "pytest",
            "pytest-cov",
            "pytest-timeout",
            "moto",
            "sphinx",
            "sphinx_rtd_theme"]}

setuptools.setup(
    name="sfini",
    version=version,
    license="MIT",
    author="Laurie",
    author_email="laurie@sitesee.com.au",
    maintainer="Laurie",
    maintainer_email="laurie@sitesee.com.au",
    description="Create, run and manage AWS Step Functions easily",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/Epic_Wink/aws-sfn-service",
    classifiers=classifiers,
    keywords="aws sfn service step functions states",
    packages=setuptools.find_packages(exclude=["tests*"]),
    python_requires="~=3.5",
    install_requires=["boto3"],
    extras_require=extras_require)



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2019/05/13

"""State definitions.

States comprise a state-machine, defining its logic and which
activities to run, and direct data.
"""

__all__ = [
    "State",
    "HasNext",
    "HasResultPath",
    "CanRetry",
    "CanCatch",
    "Succeed",
    "Fail",
    "Pass",
    "Wait",
    "Parallel",
    "Choice",
    "Task",
    "choice"]

from ._base import State
from ._base import HasNext
from ._base import HasResultPath
from ._base import CanRetry
from ._base import CanCatch
from ._state import Succeed
from ._state import Fail
from ._state import Pass
from ._state import Wait
from ._state import Parallel
from ._state import Choice
from ._state import Task
from . import choice



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2019/05/09

"""Task resource interfacing.

'Task' states require some executor to implement the task, which
different AWS services can provide, including Step Functions activities
and Lambda functions.
"""

import logging as lg

from . import _util

_logger = lg.getLogger(__name__)


class TaskResource:
    """Task execution.

    An instance of this represents a service which can run tasks defined in
    a state-machine.

    Args:
        name: name of resource
        session: session to use for AWS communication

    Attributes:
        service: name of service of which the resource belongs to
    """

    service: str = None

    def __init__(self, name: str, *, session: _util.AWSSession = None):
        self.name = name
        self.session = session

    def __str__(self):
        return "%s [%s]" % (self.name, self.service)

    __repr__ = _util.easy_repr

    @_util.cached_property
    def arn(self) -> str:
        """Task resource generated ARN."""
        region = self.session.region
        account = self.session.account_id
        fmt = "arn:aws:states:%s:%s:%s:%s"
        return fmt % (region, account, self.service, self.name)


class Lambda(TaskResource):
    """AWS Lambda function executor for a task.

    Args:
        name: name of Lambda function
        session: session to use for AWS communication
    """

    service = "function"

    @_util.cached_property
    def arn(self):
        arn_split = super().arn.split(":")
        arn_split[2] = "lambda"
        return ":".join(arn_split)



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2019/05/13

"""State definitions."""

import datetime
import typing as T
import logging as lg

from . import _base
from .. import _util

_logger = lg.getLogger(__name__)
_default = _util.DefaultParameter()


class Succeed(_base.State):
    """End execution successfully.

    Args:
        name: name of state
        comment: state description
        input_path: state input filter JSONPath, ``None`` for empty input
        output_path: state output filter JSONPath, ``None`` for discarded
            output
    """

    pass


class Fail(_base.State):
    """End execution unsuccessfully.

    Args:
        name: name of state
        comment: state description
        input_path: state input filter JSONPath, ``None`` for empty input
        output_path: state output filter JSONPath, ``None`` for discarded
            output
        error: error type
        cause: failure description
    """

    def __init__(
            self,
            name,
            comment=_default,
            input_path=_default,
            output_path=_default,
            cause: str = _default,
            error: str = _default):
        super().__init__(
            name,
            comment=comment,
            input_path=input_path,
            output_path=output_path)
        self.error = error
        self.cause = cause

    def to_dict(self):
        defn = super().to_dict()
        if self.cause != _default:
            defn["Cause"] = self.cause
        if self.error != _default:
            defn["Error"] = self.error
        return defn


class Pass(_base.HasResultPath, _base.HasNext, _base.State):
    """No-op state, possibly introducing data.

    The name specifies the location of any introduced data.

    Args:
        name: name of state
        comment: state description
        input_path: state input filter JSONPath, ``None`` for empty input
        output_path: state output filter JSONPath, ``None`` for discarded
            output
        result_path: task output location JSONPath, ``None`` for discarded
            output
        result: return value of state, stored in the variable ``name``

    Attributes:
        next: next state to execute
    """

    def __init__(
            self,
            name,
            comment=_default,
            input_path=_default,
            output_path=_default,
            result_path=_default,
            result: _util.JSONable = _default):
        super().__init__(
            name,
            comment=comment,
            input_path=input_path,
            output_path=output_path,
            result_path=result_path)
        self.result = result

    def to_dict(self):
        defn = super().to_dict()
        if self.result != _default:
            defn["Result"] = self.result
        return defn


class Wait(_base.HasNext, _base.State):
    """Wait for a time before continuing.

    Args:
        name: name of state
        until: time to wait. If ``int``, then seconds to wait; if
            ``datetime.datetime``, then time to wait until; if ``str``,
            then name of state-variable containing time to wait until
        comment: state description
        input_path: state input filter JSONPath, ``None`` for empty input
        output_path: state output filter JSONPath, ``None`` for discarded
            output

    Attributes:
        next: next state to execute
    """

    def __init__(
            self,
            name,
            until: T.Union[int, datetime.datetime, str],
            comment=_default,
            input_path=_default,
            output_path=_default):
        super().__init__(
            name,
            comment=comment,
            input_path=input_path,
            output_path=output_path)
        self.until = until

    def to_dict(self):
        defn = super().to_dict()
        if isinstance(self.until, int):
            defn["Seconds"] = self.until
        elif isinstance(self.until, datetime.datetime):
            tzinfo = self.until.tzinfo
            if tzinfo is None or tzinfo.utcoffset(self.until) is None:
                raise ValueError("Wait time must be aware")
            defn["Timestamp"] = self.until.isoformat("T")
        elif isinstance(self.until, str):
            defn["TimestampPath"] = self.until
        else:
            fmt = "Invalid type for wait time: %s"
            raise TypeError(fmt % type(self.until))
        return defn


class Parallel(
        _base.HasResultPath,
        _base.HasNext,
        _base.CanRetry,
        _base.CanCatch,
        _base.State):
    """Run states-machines in parallel.

    Args:
        name: name of state
        comment: state description
        input_path: state input filter JSONPath, ``None`` for empty input
        output_path: state output filter JSONPath, ``None`` for discarded
            output
        result_path: task output location JSONPath, ``None`` for discarded
            output

    Attributes:
        branches (list[sfini.StateMachine]): state-machines to run in
            parallel. These state-machines do not need to be registered
            with AWS Step Functions.
        next: next state to execute
        retriers: retry conditions
        catchers: handled state errors
    """

    def __init__(
            self,
            name,
            comment=_default,
            input_path=_default,
            output_path=_default,
            result_path=_default):
        super().__init__(
            name,
            comment=comment,
            input_path=input_path,
            output_path=output_path,
            result_path=result_path)
        self.branches = []

    def add(self, state_machine):
        """Add a state-machine to be executed.

        The input to the state-machine execution is the input into this
        parallel state. The output of the parallel state is a list of each
        state-machine's output (in order of adding).

        Args:
            state_machine (sfini.StateMachine): state-machine to add. It will
                be run when this task is executed. Added state-machines do not
                need to be registered with AWS Step Functions
        """

        self.branches.append(state_machine)

    def to_dict(self):
        defn = super().to_dict()
        defn["Branches"] = [sm.to_dict() for sm in self.branches]
        return defn


class Choice(_base.State):
    """Branch execution based on comparisons.

    Args:
        name: name of state
        comment: state description
        input_path: state input filter JSONPath, ``None`` for empty input
        output_path: state output filter JSONPath, ``None`` for discarded
            output

    Attributes:
        choices (list[sfini.choice.ChoiceRule]): choice rules
            determining branch conditions
        default: fall-back state if all comparisons fail, or ``None`` for
            no fall-back (Step Functions will raise a 'States.NoChoiceMatched'
            error)
    """

    def __init__(
            self,
            name,
            comment=_default,
            input_path=_default,
            output_path=_default):
        super().__init__(
            name,
            comment=comment,
            input_path=input_path,
            output_path=output_path)
        self.choices = []
        self.default: T.Union[_base.State, None] = None

    def add_to(self, states):
        super().add_to(states)
        for rule in self.choices:
            rule.next_state.add_to(states)
        if self.default is not None:
            self.default.add_to(states)

    def add(self, rule):
        """Add a choice-rule.

        Args:
            rule (sfini.choice.ChoiceRule): branch execution condition
                and specification to add

        Raises:
            RuntimeError: rule doesn't specify next-state
        """

        if rule.next_state is None:
            msg = "Top-level choice rule '%s' must specify next state"
            raise RuntimeError(msg % rule)
        self.choices.append(rule)

    def remove(self, rule):
        """Remove a branch.

        Args:
            rule (sfini.choice.ChoiceRule): branch execution condition
                and specification to remove

        Raises:
            ValueError: if rule is not a registered branch
        """

        if rule not in self.choices:
            fmt = "Rule '%s' is not registered with this state"
            raise ValueError(fmt % rule)
        self.choices.remove(rule)

    def set_default(self, state: _base.State):
        """Set the default state to execute when no conditions were met.

        Args:
            state: default state to execute
        """

        if self.default is not None:
            fmt = "Overwriting current default state '%s' with '%s'"
            _logger.warning(fmt % (self.default, state))
        self.default = state

    def to_dict(self):
        if not self.choices and self.default is None:
            raise RuntimeError("Choice '%s' has no next path")
        defn = super().to_dict()
        defn["Choices"] = [cr.to_dict() for cr in self.choices]
        if self.default is not None:
            defn["Default"] = self.default.name
        return defn


class Task(
        _base.HasResultPath,
        _base.HasNext,
        _base.CanRetry,
        _base.CanCatch,
        _base.State):
    """Activity execution.

    Args:
        name: name of state
        resource (sfini.task_resource.TaskResource): task executor, eg
            activity or Lambda function
        comment: state description
        input_path: state input filter JSONPath, ``None`` for empty input
        output_path: state output filter JSONPath, ``None`` for discarded
            output
        result_path: task output location JSONPath, ``None`` for discarded
            output
        timeout: seconds before task time-out

    Attributes:
        next: next state to execute
        retriers: retry conditions
        catchers: handled state errors
    """

    _heartbeat_extra = 5

    def __init__(
            self,
            name,
            resource,
            comment=_default,
            input_path=_default,
            output_path=_default,
            result_path=_default,
            timeout: int = _default):
        super().__init__(
            name,
            comment=comment,
            input_path=input_path,
            output_path=output_path,
            result_path=result_path)
        self.resource = resource
        self.timeout = timeout

    def to_dict(self):
        defn = super().to_dict()
        defn["Resource"] = self.resource.arn
        if self.timeout != _default:
            defn["TimeoutSeconds"] = self.timeout
        if hasattr(self.resource, "heartbeat"):
            _h = self._heartbeat_extra
            defn["HeartbeatSeconds"] = self.resource.heartbeat + _h
        return defn



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/07/11

"""AWS Step Functions service."""

import pkg_resources

try:
    __version__ = pkg_resources.get_distribution(__name__).version
except pkg_resources.DistributionNotFound:  # pragma: no cover
    __version__ = None

__all__ = [
    "AWSSession",
    "Activity",
    "ActivityRegistration",
    "CLI",
    "Lambda",
    "construct_state_machine",
    "Worker",
    "WorkerCancel"]

from ._util import AWSSession
from .activity import Activity
from .activity import ActivityRegistration
from ._cli import CLI
from .task_resource import Lambda
from .state_machine import construct_state_machine
from .worker import Worker
from .worker import WorkerCancel
from .state import Succeed
from .state import Fail
from .state import Pass
from .state import Wait
from .state import Parallel
from .state import Choice
from .state import Task
from .state.choice import And
from .state.choice import Or
from .state.choice import Not
from .state.choice import BooleanEquals
from .state.choice import NumericEquals
from .state.choice import NumericGreaterThan
from .state.choice import NumericGreaterThanEquals
from .state.choice import NumericLessThan
from .state.choice import NumericLessThanEquals
from .state.choice import StringEquals
from .state.choice import StringGreaterThan
from .state.choice import StringGreaterThanEquals
from .state.choice import StringLessThan
from .state.choice import StringLessThanEquals
from .state.choice import TimestampEquals
from .state.choice import TimestampGreaterThan
from .state.choice import TimestampGreaterThanEquals
from .state.choice import TimestampLessThan
from .state.choice import TimestampLessThanEquals

from . import activity
from . import execution
from . import state
from . import state_machine
from . import task_resource
from . import worker

import logging as lg
lg.getLogger(__name__).addHandler(lg.NullHandler())
